/*
 * This tool converts OziExplorer Track Point File Version 2.1 files into
 * Android's track file format. Android track file is a text file and
 * each line has the following structure:
 * <time> <long> <lat> <alt> [<bearing> <speed>]
 * Fields are separated by spaces and <time> is in Java's millisecond format
 * (one long number)
 */

import java.io.IOException;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.util.GregorianCalendar;

public class ozitotrack {
// Conversion factor from feet to meter
  public static final double FEETTOMETER = 0.3;
  class Location {
    double longitude = 0.0;
	double latitude = 0.0;
	double altitude = 0.0;
	long time = 0L;
  }

  public void run( String args[] ) {
	 String inpFile = null;
	 String provName = null;
     int speedFactor = 1;
	 for( int i = 0 ; i < args.length ; ++i ) {
	   String arg = args[i];
	   if( arg.startsWith( "-s" ) ) {
		 try {
		   speedFactor = Integer.parseInt( arg.substring( 2 ) );
		 } catch( NumberFormatException ex ) {
		   usage();
		   return;
	     }
	   } else {
	     if( inpFile == null )
			inpFile = arg;
         else
		 if( provName == null )
			provName = arg;
		 else {
		   usage();
		   return;
		 }
	   }
	 }
	 if( inpFile == null || provName == null ) {
	   usage();
	   return;
	 }
	 BufferedReader inf = null;
	 try {
	   FileReader f = new FileReader( inpFile );
       inf = new BufferedReader( f );
     } catch( IOException ex ) {
	   System.out.println( "File not found: "+inpFile );
	   return;
	 }
	 String outFileName = provName +"_track";
	 String locFileName = provName +"_location";
	 try {
       previousTimeStamp = -1;
	   previousTimeDiff = -1L;
	   previousCalculatedTimeStamp = -1L;
	   long trackStart = -1;
	   long trackEnd = -1;
	   FileWriter f = new FileWriter( outFileName );	   
	   PrintWriter outf = new PrintWriter( f );
	   FileWriter lf = new FileWriter( locFileName );
	   PrintWriter outlf = new PrintWriter( lf );
	   boolean firstLocation = true;
	   do {
		 String line = inf.readLine();
		 if( line == null )
			break;
		 Location l = parseOziLine( line,speedFactor );
		 if( l == null )
			continue;
		 if( firstLocation ) {
		   trackStart = l.time;
		   writeLocation( outlf,provName,l );
		   firstLocation = false;
		 }
		 trackEnd = l.time;
		 l.time -= trackStart;
		 writeTrackLine( outf, l );
	   } while( true );
	   f.close();
	   lf.close();
	   double trackLength = (double)( trackEnd - trackStart ) / 60000.0;
	   System.out.println( "Track length: "+trackLength+" minutes ("+
						(trackLength / 60.0)+" hours approx.)" );
     } catch( IOException ex ) {
	   ex.printStackTrace();
	 }
  }

  public static void main( String args[] ) {
	ozitotrack inst = new ozitotrack();
	inst.run( args );
  }

  private long firstTimeStamp = -1;
  private long previousTimeStamp = -1L;
  private long previousTimeDiff = -1L;
  private long previousCalculatedTimeStamp = -1L;

  private Location parseOziLine( String line,int speedFactor ) {
	StringTokenizer st = new StringTokenizer( line,"\t ," );
	double longitude = 0.0;
    double latitude = 0.0;
	double altitude = 0.0;
	int day = 0;
    int month = 0;
	int year = 0;
	int hour = 0;
	int minute = 0;
	int second = 0;
	long timeStamp = 0L;
	Location l = null;
	try {
	  latitude = Double.parseDouble( st.nextToken() );
	  longitude = Double.parseDouble( st.nextToken() );
	  st.nextToken();		// skip waypoint flag
	  altitude = Double.parseDouble( st.nextToken() ) * FEETTOMETER;
	  String timeField = st.nextToken();
	  String dateField = st.nextToken();
	  String hourField = st.nextToken();
	  if( ( dateField.length() == 0 ) ||
		  ( hourField.length() == 0 ) )
		return null;
	  StringTokenizer dateTokenizer = new StringTokenizer( dateField,"-" );
	  day = Integer.parseInt( dateTokenizer.nextToken() );
	  String monthName = dateTokenizer.nextToken();
	  month = searchMonth( monthName );
	  if( month == -1 )
		return null;
	  year = Integer.parseInt( dateTokenizer.nextToken() )+2000;
	  StringTokenizer hourTokenizer = new StringTokenizer( hourField,":" );
	  hour = Integer.parseInt( hourTokenizer.nextToken() );
	  minute = Integer.parseInt( hourTokenizer.nextToken() );
	  second = Integer.parseInt( hourTokenizer.nextToken() );
	  l = new Location();
	  GregorianCalendar gc = new GregorianCalendar( year,month,day,hour,minute,second );
	  l.longitude = longitude;
	  l.latitude = latitude;
	  l.altitude = altitude;
// apply the speed correction
	  long nextTimeStamp = gc.getTimeInMillis();
	  long timeDiff = -1;
	  long calculatedTimeStamp = -1;
	  if( previousTimeStamp != -1 ) {
		timeDiff = nextTimeStamp - previousTimeStamp;
		if( timeDiff < 0 )
			timeDiff = -timeDiff;
		timeDiff /= speedFactor;
// There are unexplanaible time jumps in the input file - smooth it
		if( ( previousTimeDiff >= 0 ) &&
			( timeDiff > previousTimeDiff*10 ) )
			timeDiff = previousTimeDiff;
		calculatedTimeStamp = previousCalculatedTimeStamp + timeDiff;
      } else
		firstTimeStamp = calculatedTimeStamp = nextTimeStamp;
	  l.time = calculatedTimeStamp;
	  previousCalculatedTimeStamp = calculatedTimeStamp;
	  previousTimeStamp = nextTimeStamp;
	  previousTimeDiff = timeDiff;
	} catch( NoSuchElementException ex ) {
	  return null;
	} catch( NumberFormatException ex ) {
	  return null;
	}
	return l;
  }

  private void writeTrackLine( PrintWriter outf, Location l ) throws IOException {
    outf.println( l.time+" "+l.longitude+" "+l.latitude+" "+l.altitude );
  }

  private void writeLocation( PrintWriter outlf, String provName, Location l ) {
	outlf.println( provName+","+l.time+","+l.longitude+","+l.latitude+
					","+l.altitude+",0.0,0.0" );
  }

  String monthNames[] = { 
	"jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"
  };

  private int searchMonth( String monthName ) {
	for( int i = 0 ; i < monthNames.length ; ++i ) {
	  if( monthNames[i].equals( monthName ) )
		return i;
	}
	return -1;
  }

  private void usage() {
	System.out.println( "Usage: java ozitrack [-s<speedfactor>] <ozi file> <provname>" );
	System.out.println( "  where <provname> is the name of the provider (<provname>_location and <provname>_track will be generated)" );
    System.out.println( "  <speedfactor> is used to divide time period lengths among the fixes, 1 by default" );
  }
}
